
#ifndef CONSTANTS_H
#define CONSTANTS_H

extern const double epsilon;
extern const double splitter;

extern const double resulterrbound;

extern const double ccwerrboundA;
extern const double ccwerrboundB;
extern const double ccwerrboundC;

extern const double o3derrboundA;
extern const double o3derrboundB;
extern const double o3derrboundC;

extern const double iccerrboundA;
extern const double iccerrboundB;
extern const double iccerrboundC;

extern const double isperrboundA;
extern const double isperrboundB;
extern const double isperrboundC;

#endif
